@@ -1,11 +1,12 @@
 import collections
-import torch
-from torch import fx
-from e3nn import o3
-from e3nn.math import germinate_formulas, reduce_permutation
-from e3nn.util import explicit_default_types
-from e3nn.util.jit import compile_mode
 
+import torch
+from e3nn import o3
+from e3nn.math import germinate_formulas, orthonormalize, reduce_permutation
+from e3nn.util import explicit_default_types
+from e3nn.util.codegen import CodeGenMixin
+from e3nn.util.jit import compile_mode
+from torch import fx
 
 _TP = collections.namedtuple('tp', 'op, args')
 _INPUT = collections.namedtuple('input', 'tensor, start, stop')
@@ -75,8 +76,8 @@ def _get_ops(path):
         yield op
 
 
-@compile_mode('script')
-class ReducedTensorProducts(fx.GraphModule):
+@compile_mode('trace')
+class ReducedTensorProducts(CodeGenMixin, torch.nn.Module):
     r"""reduce a tensor with symmetries into irreducible representations
 
     Parameters
@@ -128,10 +129,19 @@ class ReducedTensorProducts(fx.GraphModule):
     # pylint: disable=abstract-method
 
     def __init__(self, formula, filter_ir_out=None, filter_ir_mid=None, eps=1e-9, **irreps):
-        super().__init__(self, fx.Graph())
+        super().__init__()
 
         if filter_ir_out is not None:
-            filter_ir_out = [o3.Irrep(ir) for ir in filter_ir_out]
+            try:
+                filter_ir_out = [o3.Irrep(ir) for ir in filter_ir_out]
+            except ValueError:
+                raise ValueError(f"filter_ir_out (={filter_ir_out}) must be an iterable of e3nn.o3.Irrep")
+
+        if filter_ir_mid is not None:
+            try:
+                filter_ir_mid = [o3.Irrep(ir) for ir in filter_ir_mid]
+            except ValueError:
+                raise ValueError(f"filter_ir_mid (={filter_ir_mid}) must be an iterable of e3nn.o3.Irrep")
 
         f0, formulas = germinate_formulas(formula)
 
@@ -170,26 +180,28 @@ class ReducedTensorProducts(fx.GraphModule):
 
         for ir, path, base_o3 in _wigner_nj(*[irreps[i] for i in f0], filter_ir_mid=filter_ir_mid, dtype=torch.float64):
             if filter_ir_out is None or ir in filter_ir_out:
-                P = base_o3.flatten(1) @ base_perm.flatten(1).T
-                if P.norm() > eps:  # if this Irrep is present in the premutation basis we keep it
-                    Ps[ir].append((path, base_o3))
+                # P = base_o3.flatten(1) @ base_perm.flatten(1).T
+                # if P.norm() > eps:  # if this Irrep is present in the premutation basis we keep it
+                Ps[ir].append((path, base_o3))
 
         outputs = []
         change_of_basis = []
         irreps_out = []
 
+        P = base_perm.flatten(1)  # [permutation basis, input basis] (a,omega)
+        PP = P @ P.T  # (a,a)
+
         for ir in Ps:
             mul = len(Ps[ir])
             paths = [path for path, _ in Ps[ir]]
             base_o3 = torch.stack([R for _, R in Ps[ir]])
 
             R = base_o3.flatten(2)  # [multiplicity, ir, input basis] (u,j,omega)
-            P = base_perm.flatten(1)  # [permutation basis, input basis] (a,omega)
 
-            Xs = []
+            proj_s = []  # list of projectors into vector space
             for j in range(ir.dim):
+                # Solve X @ R[:, j] = Y @ P, but keep only X
                 RR = R[:, j] @ R[:, j].T  # (u,u)
-                PP = P @ P.T  # (a,a)
                 RP = R[:, j] @ P.T  # (u,a)
 
                 prob = torch.cat([
@@ -198,24 +210,16 @@ class ReducedTensorProducts(fx.GraphModule):
                 ], dim=0)
                 eigenvalues, eigenvectors = torch.linalg.eigh(prob)
                 X = eigenvectors[:, eigenvalues < eps][:mul].T  # [solutions, multiplicity]
-                X = torch.linalg.qr(X, mode='r').R
-                for i, x in enumerate(X):
-                    for j in range(i, mul):
-                        if x[j] < eps:
-                            x.neg_()
-                        if x[j] > eps:
-                            break
+                proj_s.append(X.T @ X)
 
-                X[X.abs() < eps] = 0
-                X = sorted([[x.item() for x in line] for line in X])
-                X = torch.tensor(X, dtype=torch.float64)
+                break  # do not check all components because too time expensive
 
-                Xs.append(X)
+            for p in proj_s:
+                assert (p - proj_s[0]).abs().max() < eps, f"found different solutions for irrep {ir}"
 
-            for X in Xs:
-                assert (X - Xs[0]).abs().max() < eps
+            # look for an X such that X.T @ X = Projector
+            X, _ = orthonormalize(proj_s[0], eps)
 
-            X = Xs[0]
             for x in X:
                 C = torch.einsum("u,ui...->i...", x, base_o3)
                 correction = (ir.dim / C.pow(2).sum())**0.5
@@ -234,10 +238,12 @@ class ReducedTensorProducts(fx.GraphModule):
                 for op in _get_ops(p):
                     tps.add(op)
 
+        root = torch.nn.Module()
+
         tps = list(tps)
         for i, op in enumerate(tps):
             tp = o3.TensorProduct(op[0], op[1], op[2], [(0, 0, 0, 'uuu', False)])
-            setattr(self, f'tp{i}', tp)
+            setattr(root, f'tp{i}', tp)
 
         graph = fx.Graph()
         tracer = torch.fx.proxy.GraphAppendingTracer(graph)
@@ -281,9 +287,11 @@ class ReducedTensorProducts(fx.GraphModule):
 
         out = torch.cat(outs, dim=-1)
         graph.output(out.node)
+        graphmod = fx.GraphModule(root, graph, "main")
 
-        self.graph = graph
-        self.recompile()
+        self._codegen_register({
+            "main": graphmod
+        })
 
     def __repr__(self):
         return (
@@ -292,3 +300,6 @@ class ReducedTensorProducts(fx.GraphModule):
             f"    out: {self.irreps_out}\n"
             ")"
         )
+
+    def forward(self, *xs):
+        return self.main(*xs)
