@@ -1,11 +1,7 @@
 from ._instruction import Instruction
-from ._tensor_product import (
-    TensorProduct,
-    FullyConnectedTensorProduct,
-    ElementwiseTensorProduct,
-    FullTensorProduct,
-)
-
+from ._tensor_product import TensorProduct
+from ._sub import (ElementwiseTensorProduct, FullTensorProduct,
+                   FullyConnectedTensorProduct, TensorSquare)
 
 __all__ = [
     "Instruction",
@@ -13,4 +9,5 @@ __all__ = [
     "FullyConnectedTensorProduct",
     "ElementwiseTensorProduct",
     "FullTensorProduct",
+    "TensorSquare",
 ]
