@@ -1,6 +1,6 @@
 import warnings
 from math import sqrt
-from typing import Iterator, List, Optional, Union
+from typing import List, Optional, Union, Any
 
 import e3nn
 import torch
@@ -171,16 +171,13 @@ class TensorProduct(CodeGenMixin, torch.nn.Module):
     >>> assert vars.min() > 1 / 3
     >>> assert vars.max() < 3
     """
-    _specialized_code: bool
-    _optimize_einsums: bool
-    _profiling_str: str
-    normalization: str
+    instructions: List[Any]
     shared_weights: bool
     internal_weights: bool
     weight_numel: int
-    in1_var: List[float]
-    in2_var: List[float]
-    out_var: List[float]
+    _specialized_code: bool
+    _optimize_einsums: bool
+    _profiling_str: str
     _in1_dim: int
     _in2_dim: int
 
@@ -193,8 +190,8 @@ class TensorProduct(CodeGenMixin, torch.nn.Module):
         in1_var: Optional[Union[List[float], torch.Tensor]] = None,
         in2_var: Optional[Union[List[float], torch.Tensor]] = None,
         out_var: Optional[Union[List[float], torch.Tensor]] = None,
-        irrep_normalization: str = 'component',
-        path_normalization: str = 'element',
+        irrep_normalization: str = None,
+        path_normalization: str = None,
         internal_weights: Optional[bool] = None,
         shared_weights: Optional[bool] = None,
         compile_left_right: bool = True,
@@ -213,8 +210,14 @@ class TensorProduct(CodeGenMixin, torch.nn.Module):
             )
             irrep_normalization = normalization
 
-        assert irrep_normalization in ['component', 'norm']
-        assert path_normalization in ['element', 'path']
+        if irrep_normalization is None:
+            irrep_normalization = 'component'
+
+        if path_normalization is None:
+            path_normalization = 'element'
+
+        assert irrep_normalization in ['component', 'norm', 'none']
+        assert path_normalization in ['element', 'path', 'none']
 
         self.irreps_in1 = o3.Irreps(irreps_in1)
         self.irreps_in2 = o3.Irreps(irreps_in2)
@@ -224,14 +227,21 @@ class TensorProduct(CodeGenMixin, torch.nn.Module):
         instructions = [x if len(x) == 6 else x + (1.0,) for x in instructions]
         instructions = [
             Instruction(
-                i_in1, i_in2, i_out, connection_mode, has_weight, path_weight,
-                {
+                i_in1=i_in1,
+                i_in2=i_in2,
+                i_out=i_out,
+                connection_mode=connection_mode,
+                has_weight=has_weight,
+                path_weight=path_weight,
+                path_shape={
                     'uvw': (self.irreps_in1[i_in1].mul, self.irreps_in2[i_in2].mul, self.irreps_out[i_out].mul),
                     'uvu': (self.irreps_in1[i_in1].mul, self.irreps_in2[i_in2].mul),
                     'uvv': (self.irreps_in1[i_in1].mul, self.irreps_in2[i_in2].mul),
                     'uuw': (self.irreps_in1[i_in1].mul, self.irreps_out[i_out].mul),
                     'uuu': (self.irreps_in1[i_in1].mul,),
                     'uvuv': (self.irreps_in1[i_in1].mul, self.irreps_in2[i_in2].mul),
+                    'uvu<v': (self.irreps_in1[i_in1].mul * (self.irreps_in2[i_in2].mul - 1) // 2,),
+                    'u<vw': (self.irreps_in1[i_in1].mul * (self.irreps_in2[i_in2].mul - 1) // 2, self.irreps_out[i_out].mul),
                 }[connection_mode],
             )
             for i_in1, i_in2, i_out, connection_mode, has_weight, path_weight in instructions
@@ -263,6 +273,8 @@ class TensorProduct(CodeGenMixin, torch.nn.Module):
                 'uuw': self.irreps_in1[ins.i_in1].mul,
                 'uuu': 1,
                 'uvuv': 1,
+                'uvu<v': 1,
+                'u<vw': self.irreps_in1[ins.i_in1].mul * (self.irreps_in2[ins.i_in2].mul - 1) // 2,
             }[ins.connection_mode]
 
         normalization_coefficients = []
@@ -272,14 +284,14 @@ class TensorProduct(CodeGenMixin, torch.nn.Module):
             mul_ir_out = self.irreps_out[ins.i_out]
             assert mul_ir_in1.ir.p * mul_ir_in2.ir.p == mul_ir_out.ir.p
             assert abs(mul_ir_in1.ir.l - mul_ir_in2.ir.l) <= mul_ir_out.ir.l <= mul_ir_in1.ir.l + mul_ir_in2.ir.l
-            assert ins.connection_mode in ['uvw', 'uvu', 'uvv', 'uuw', 'uuu', 'uvuv']
-
-            alpha = 1
+            assert ins.connection_mode in ['uvw', 'uvu', 'uvv', 'uuw', 'uuu', 'uvuv', 'uvu<v', 'u<vw']
 
             if irrep_normalization == 'component':
-                alpha *= mul_ir_out.ir.dim
+                alpha = mul_ir_out.ir.dim
             if irrep_normalization == 'norm':
-                alpha *= mul_ir_in1.ir.dim * mul_ir_in2.ir.dim
+                alpha = mul_ir_in1.ir.dim * mul_ir_in2.ir.dim
+            if irrep_normalization == 'none':
+                alpha = 1
 
             if path_normalization == 'element':
                 x = sum(
@@ -290,6 +302,8 @@ class TensorProduct(CodeGenMixin, torch.nn.Module):
             if path_normalization == 'path':
                 x = in1_var[ins.i_in1] * in2_var[ins.i_in2] * num_elements(ins)
                 x *= len([i for i in instructions if i.i_out == ins.i_out])
+            if path_normalization == 'none':
+                x = 1
 
             if x > 0.0:
                 alpha /= x
@@ -706,7 +720,7 @@ class TensorProduct(CodeGenMixin, torch.nn.Module):
             codes += [Path.MOVETO, Path.LINETO]
 
             if plot_weight:
-                color = cmap(path_weight[ins_index]) if ins.has_weight else 'black'
+                color = cmap(0.5 + 0.5 * path_weight[ins_index]) if ins.has_weight else 'black'
             else:
                 color = 'green' if ins.has_weight else 'black'
 
@@ -716,7 +730,7 @@ class TensorProduct(CodeGenMixin, torch.nn.Module):
                 edgecolor=color,
                 alpha=0.5,
                 ls='-',
-                lw=1.5 * ins.path_weight / min(i.path_weight for i in self.instructions),
+                lw=1.5,
             ))
 
         # add labels
@@ -766,201 +780,3 @@ class TensorProduct(CodeGenMixin, torch.nn.Module):
         ax.axis('off')
 
         return fig, ax
-
-
-class FullyConnectedTensorProduct(TensorProduct):
-    r"""Fully-connected weighted tensor product
-
-    All the possible path allowed by :math:`|l_1 - l_2| \leq l_{out} \leq l_1 + l_2` are made.
-    The output is a sum on different paths:
-
-    .. math::
-
-        z_w = \sum_{u,v} w_{uvw} x_u \otimes y_v + \cdots \text{other paths}
-
-    where :math:`u,v,w` are the indices of the multiplicities.
-
-    Parameters
-    ----------
-    irreps_in1 : `e3nn.o3.Irreps`
-        representation of the first input
-
-    irreps_in2 : `e3nn.o3.Irreps`
-        representation of the second input
-
-    irreps_out : `e3nn.o3.Irreps`
-        representation of the output
-
-    normalization : {'component', 'norm'}
-        see `e3nn.o3.TensorProduct`
-
-    internal_weights : bool
-        see `e3nn.o3.TensorProduct`
-
-    shared_weights : bool
-        see `e3nn.o3.TensorProduct`
-    """
-    def __init__(
-        self,
-        irreps_in1,
-        irreps_in2,
-        irreps_out,
-        **kwargs
-    ):
-        irreps_in1 = o3.Irreps(irreps_in1)
-        irreps_in2 = o3.Irreps(irreps_in2)
-        irreps_out = o3.Irreps(irreps_out)
-
-        instr = [
-            (i_1, i_2, i_out, 'uvw', True, 1.0)
-            for i_1, (_, ir_1) in enumerate(irreps_in1)
-            for i_2, (_, ir_2) in enumerate(irreps_in2)
-            for i_out, (_, ir_out) in enumerate(irreps_out)
-            if ir_out in ir_1 * ir_2
-        ]
-        super().__init__(irreps_in1, irreps_in2, irreps_out, instr, **kwargs)
-
-
-class ElementwiseTensorProduct(TensorProduct):
-    r"""Elementwise connected tensor product.
-
-    .. math::
-
-        z_u = x_u \otimes y_u
-
-    where :math:`u` runs over the irreps. Note that there are no weights.
-    The output representation is determined by the two input representations.
-
-    Parameters
-    ----------
-    irreps_in1 : `e3nn.o3.Irreps`
-        representation of the first input
-
-    irreps_in2 : `e3nn.o3.Irreps`
-        representation of the second input
-
-    filter_ir_out : iterator of `e3nn.o3.Irrep`, optional
-        filter to select only specific `e3nn.o3.Irrep` of the output
-
-    normalization : {'component', 'norm'}
-        see `e3nn.o3.TensorProduct`
-
-    Examples
-    --------
-    Elementwise scalar product
-
-    >>> ElementwiseTensorProduct("5x1o + 5x1e", "10x1e", ["0e", "0o"])
-    ElementwiseTensorProduct(5x1o+5x1e x 10x1e -> 5x0o+5x0e | 10 paths | 0 weights)
-
-    """
-    def __init__(
-        self,
-        irreps_in1,
-        irreps_in2,
-        filter_ir_out=None,
-        **kwargs
-    ):
-
-        irreps_in1 = o3.Irreps(irreps_in1).simplify()
-        irreps_in2 = o3.Irreps(irreps_in2).simplify()
-        if filter_ir_out is not None:
-            filter_ir_out = [o3.Irrep(ir) for ir in filter_ir_out]
-
-        assert irreps_in1.num_irreps == irreps_in2.num_irreps
-
-        irreps_in1 = list(irreps_in1)
-        irreps_in2 = list(irreps_in2)
-
-        i = 0
-        while i < len(irreps_in1):
-            mul_1, ir_1 = irreps_in1[i]
-            mul_2, ir_2 = irreps_in2[i]
-
-            if mul_1 < mul_2:
-                irreps_in2[i] = (mul_1, ir_2)
-                irreps_in2.insert(i + 1, (mul_2 - mul_1, ir_2))
-
-            if mul_2 < mul_1:
-                irreps_in1[i] = (mul_2, ir_1)
-                irreps_in1.insert(i + 1, (mul_1 - mul_2, ir_1))
-            i += 1
-
-        out = []
-        instr = []
-        for i, ((mul, ir_1), (mul_2, ir_2)) in enumerate(zip(irreps_in1, irreps_in2)):
-            assert mul == mul_2
-            for ir in ir_1 * ir_2:
-
-                if filter_ir_out is not None and ir not in filter_ir_out:
-                    continue
-
-                i_out = len(out)
-                out.append((mul, ir))
-                instr += [
-                    (i, i, i_out, 'uuu', False)
-                ]
-
-        super().__init__(irreps_in1, irreps_in2, out, instr, **kwargs)
-
-
-class FullTensorProduct(TensorProduct):
-    r"""Full tensor product between two irreps.
-
-    .. math::
-
-        z_{uv} = x_u \otimes y_v
-
-    where :math:`u` and :math:`v` run over the irreps. Note that there are no weights.
-    The output representation is determined by the two input representations.
-
-    Parameters
-    ----------
-    irreps_in1 : `e3nn.o3.Irreps`
-        representation of the first input
-
-    irreps_in2 : `e3nn.o3.Irreps`
-        representation of the second input
-
-    filter_ir_out : iterator of `e3nn.o3.Irrep`, optional
-        filter to select only specific `e3nn.o3.Irrep` of the output
-
-    normalization : {'component', 'norm'}
-        see `e3nn.o3.TensorProduct`
-    """
-    def __init__(
-        self,
-        irreps_in1: o3.Irreps,
-        irreps_in2: o3.Irreps,
-        filter_ir_out: Iterator[o3.Irrep] = None,
-        **kwargs
-    ):
-
-        irreps_in1 = o3.Irreps(irreps_in1).simplify()
-        irreps_in2 = o3.Irreps(irreps_in2).simplify()
-        if filter_ir_out is not None:
-            filter_ir_out = [o3.Irrep(ir) for ir in filter_ir_out]
-
-        out = []
-        instr = []
-        for i_1, (mul_1, ir_1) in enumerate(irreps_in1):
-            for i_2, (mul_2, ir_2) in enumerate(irreps_in2):
-                for ir_out in ir_1 * ir_2:
-
-                    if filter_ir_out is not None and ir_out not in filter_ir_out:
-                        continue
-
-                    i_out = len(out)
-                    out.append((mul_1 * mul_2, ir_out))
-                    instr += [
-                        (i_1, i_2, i_out, 'uvuv', False)
-                    ]
-
-        out = o3.Irreps(out)
-        out, p, _ = out.sort()
-
-        instr = [
-            (i_1, i_2, p[i_out], mode, train)
-            for i_1, i_2, i_out, mode, train in instr
-        ]
-
-        super().__init__(irreps_in1, irreps_in2, out, instr, **kwargs)
